import os
import importlib
from datetime import datetime
from pathlib import Path
from bs4 import BeautifulSoup
from joblib import Parallel, delayed
if importlib.util.find_spec("pandas") is not None:
    import pandas as pd
from xbrr.base.reader.base_reader import BaseReader
from xbrr.edinet.reader.directory import Directory
from xbrr.edinet.reader.taxonomy import Taxonomy
from xbrr.edinet.reader.element import Element
from xbrr.edinet.reader.element_schema import ElementSchema


class Reader(BaseReader):

    def __init__(self, xbrl_dir_or_file="", taxonomy=None):
        super().__init__("edinet")
        self.__xbrl_dir_or_file = xbrl_dir_or_file
        self._cache = {}

        if os.path.isdir(xbrl_dir_or_file):
            self.xbrl_dir = Directory(xbrl_dir_or_file)
            self.xbrl_file = self.xbrl_dir._find_file("xbrl", as_xml=False)
        elif os.path.isfile(xbrl_dir_or_file):
            self.xbrl_file = xbrl_dir_or_file
            self.xbrl_dir = None
        else:
            raise Exception(
                f"File or directory {xbrl_dir_or_file} does not Exsit.")

        if isinstance(taxonomy, Taxonomy):
            self.taxonomy = taxonomy
        else:
            if self.xbrl_dir:
                root = Path(self.xbrl_dir.root).parent
            else:
                root = Path(self.xbrl_file).parent

            if root.name == "raw":
                # Cookiecutter data science structure
                root = root.parent
            root = root.joinpath("external")
            self.taxonomy = Taxonomy(root)
        self.taxonomy_year = ""
        self.__set_taxonomy_year()

    def set_cache(self, cache):
        self._cache = cache
        return self

    def __reduce_ex__(self, proto):
        return type(self), (self.__xbrl_dir_or_file, self.taxonomy)

    def __set_taxonomy_year(self):
        self.taxonomy_year = ""
        date = self.xbrl.find("jpdei_cor:CurrentFiscalYearEndDateDEI").text
        kind = self.xbrl.find("jpdei_cor:TypeOfCurrentPeriodDEI").text
        date = datetime.strptime(date, "%Y-%m-%d")
        for y in sorted(list(self.taxonomy.TAXONOMIES.keys()), reverse=True):
            boarder_date = datetime(int(y[:4]), 3, 31)
            if kind[0] in ("Q", "H") and date > boarder_date:
                self.taxonomy_year = y
            elif date >= boarder_date:
                if y == 2019:
                    self.taxonomy_year = "2019_cg_ifrs"
                else:
                    self.taxonomy_year = y

            if self.taxonomy_year:
                break

    @property
    def roles(self):
        role_refs = self.find_all("link:roleRef")
        roles = {}
        for e in role_refs:
            element = e.element
            link = element["xlink:href"]
            roles[element["roleURI"]] = {
                "link": element["xlink:href"],
                "name": self.read_by_link(link).element.find("link:definition").text
            }

        return roles

    @property
    def taxonomy_path(self):
        return self.taxonomy.root.joinpath("taxonomy", str(self.taxonomy_year))

    @property
    def namespaces(self):
        schema = self.xbrl.find("xbrli:xbrl")
        namespaces = {}
        for a in schema.attrs:
            if a.startswith("xmlns:"):
                namespaces[a.replace("xmlns:", "")] = schema.attrs[a]

        return namespaces

    @property
    def xbrl(self):
        if self.xbrl_dir:
            path = self.xbrl_dir._find_file("xbrl", as_xml=False)
        else:
            path = self.xbrl_file
        return self._read_from_cache(path)

    def _read_from_cache(self, path):
        xml = None
        if path in self._cache:
            xml = self._cache[path]
        else:
            with open(path, encoding="utf-8-sig") as f:
                xml = BeautifulSoup(f, "lxml-xml")
            self._cache[path] = xml
        return self._cache[path]

    def link_to_path(self, link):
        path = link
        element = ""
        if "#" in link:
            path, element = link.split("#")
        if self.taxonomy and path.startswith(self.taxonomy.prefix):
            path = path.replace(self.taxonomy.prefix, "")
            path = os.path.join(self.taxonomy_path, path)
            if not os.path.exists(path):
                _path = Path(path)
                xbrl_date = _path.parent.name
                # check namespace directory
                taxonomy_date = ""
                if _path.parent.parent.exists():
                    for d in _path.parent.parent.iterdir():
                        if d.is_dir():
                            taxonomy_date = d.name
                            break

                if taxonomy_date and taxonomy_date != xbrl_date:
                    path = path.replace(xbrl_date, taxonomy_date)
            if os.path.isdir(path):
                _path = Path(path)
                xbrl_date = _path.name
                # element should exist if name does not exist.
                namespace = "_".join(element.split("_")[:-1])
                path = _path.joinpath(f"{namespace}_{xbrl_date}.xsd")
        elif self.xbrl_dir:
            path = self.xbrl_dir._find_file("xsd", as_xml=False)
        else:
            path = os.path.dirname(self.xbrl_file)

        return path

    def read_by_link(self, link):
        if link.startswith(self.taxonomy.prefix):
            self.taxonomy.download(self.taxonomy_year)

        element = ""
        if "#" in link:
            element = link.split("#")[-1]

        path = self.link_to_path(link)
        xml = self._read_from_cache(path)

        if element:
            xml = xml.select(f"#{element}")
            # xml = xml.find("element", {"id": element})
            if len(xml) > 0:
                xml = xml[0]
            xml = Element(element, xml, link, self)

        return xml

    def has_role_in_link(self, role_link, link_type):
        if link_type == "presentation":
            doc = self.xbrl_dir.pre
        elif link_type == "calculation":
            doc = self.xbrl_dir.cal
        else:
            return False

        role = doc.find("link:roleRef", {"roleURI": role_link})
        if role is not None:
            return True
        else:
            return False

    def read_schema_by_role(self, role_link, link_type="presentation",
                            label_kind="", label_verbose=False):
        if self.xbrl_dir is None:
            raise Exception("XBRL directory is required.")

        doc = None
        link_node = ""
        arc_node = ""
        if link_type == "presentation":
            doc = self.xbrl_dir.pre
            link_node = "link:presentationLink"
            arc_node = "link:presentationArc"
        elif link_type == "calculation":
            doc = self.xbrl_dir.cal
            link_node = "link:calculationLink"
            arc_node = "link:calculationArc"
        else:
            raise Exception(f"Does not support {link_type}.")

        schemas = []
        role = doc.find(link_node, {"xlink:role": role_link})
        if role is None:
            return schemas

        def get_name(loc):
            return loc["xlink:href"].split("#")[-1]

        def create(reader, reference):
            return ElementSchema.create_from_reference(
                        reader, reference, label_kind, label_verbose)

        nodes = {}
        arc_role = ""
        if link_type == "calculation":
            arc_role = "summation-item"
        else:
            arc_role = "parent-child"

        locs = {}
        for loc in role.find_all("link:loc"):
            locs[loc["xlink:label"]] = loc

        for i, arc in enumerate(role.find_all(arc_node)):
            if not arc["xlink:arcrole"].endswith(arc_role):
                print("Unexpected arctype.")
                continue

            parent = locs[arc["xlink:from"]]
            child = locs[arc["xlink:to"]]

            if get_name(child) not in nodes:
                c = create(self, child["xlink:href"]).set_alias(child["xlink:label"])
                nodes[get_name(child)] = Node(c, arc["order"])
            else:
                nodes[get_name(child)].order = arc["order"]

            if get_name(parent) not in nodes:
                p = create(self, parent["xlink:href"]).set_alias(parent["xlink:label"])
                nodes[get_name(parent)] = Node(p, i)

            nodes[get_name(child)].add_parent(nodes[get_name(parent)])

        parent_depth = -1
        for name in nodes:
            if parent_depth < nodes[name].depth:
                parent_depth = nodes[name].depth

        for name in nodes:
            n = nodes[name]
            item = {}
            parents = n.get_parents()
            parents = parents + ([""] * (parent_depth - len(parents)))

            for i, p in zip(reversed(range(parent_depth)), parents):
                name = p if isinstance(p, str) else p.name
                order = "0" if isinstance(p, str) else p.order
                item[f"parent_{i}"] = name
                item[f"parent_{i}_label"] = ""
                item[f"parent_{i}_order"] = order

            item["order"] = n.order
            item["depth"] = n.depth
            item.update(n.element.to_dict())
            schemas.append(item)

        schemas = pd.DataFrame(schemas)
        schemas.sort_values(by=[c for c in schemas.columns
                                if c.endswith("order")],
                            inplace=True)

        label_dict = pd.Series(schemas["label"].tolist(),
                               index=schemas["name"].tolist()).to_dict()

        for i, row in schemas.iterrows():
            for j in range(parent_depth):
                name = row[f"parent_{j}"]
                if name in label_dict:
                    schemas.loc[i, f"parent_{j}_label"] = label_dict[name]

        return schemas

    def read_value_by_role(self, role_link, link_type="presentation",
                           label_kind="", label_verbose=False):

        schemas = self.read_schema_by_role(role_link, link_type,
                                           label_kind, label_verbose)

        targets = []
        xbrl_data = []
        if len(schemas) == 0:
            return None

        for i, row in schemas.iterrows():
            tag_name = row["name"]

            for n in self.namespaces:
                if tag_name.startswith(n):
                    tag_name = tag_name.replace(f"{n}_", f"{n}:")
                    break

            tag_element = self.xbrl.find(tag_name)
            if tag_element is None:
                continue

            item = {}
            for k in schemas.columns:
                item[k] = row[k]

            targets.append((tag_name, item))

        def read_value(reader, target, label_kind, label_verbose):
            name, item = target
            results = []
            for element in reader.find_all(name):
                value = element.value(label_kind=None,
                                      label_verbose=False).to_dict()
                _item = {}
                for k in value:
                    if k in item:
                        _item[k] = item[k]
                    else:
                        _item[k] = value[k]

                results.append(_item)
            return results

        results = Parallel(n_jobs=-1)(delayed(read_value)(
                                        self, t, label_kind, label_verbose)
                                      for t in targets)

        for r in results:
            if len(r) > 0:
                xbrl_data += r

        xbrl_data = pd.DataFrame(xbrl_data)
        return xbrl_data

    def find(self, tag, attrs={}, recursive=True, text=None,
             **kwargs):
        element = self.xbrl.find(tag, attrs, recursive, text, **kwargs)
        return self._to_element(tag, element)

    def find_all(self, tag, attrs={}, recursive=True, text=None,
                 limit=None, **kwargs):
        elements = self.xbrl.find_all(
                        tag, attrs, recursive, text, limit, **kwargs)

        return [self._to_element(tag, e) for e in elements]

    def _to_element(self, tag, element):
        if element is None:
            return None

        reference = tag.replace(":", "_")
        if element.namespace:
            path = self.link_to_path(element.namespace)
            namespace_root, prefix = os.path.split(path)
            xsd_name = ""
            if os.path.isdir(namespace_root):
                fs = [f for f in os.listdir(namespace_root)
                      if f.startswith(prefix) and f.endswith(".xsd")]
                if len(fs) > 0:
                    xsd_name = fs[0]

            if element.namespace.startswith(self.taxonomy.prefix):
                namespace_root = namespace_root.replace(
                                    str(self.taxonomy_path) + os.sep,
                                    self.taxonomy.prefix)
            else:
                namespace_root = ""

            if namespace_root:
                reference = f"{namespace_root}/{xsd_name}#{reference}"
            else:
                reference = f"{xsd_name}#{reference}"

        return Element(tag, element, reference, self)


class Node():

    def __init__(self, element, order=0):
        self.element = element
        self.parent = None
        self.order = order

    def add_parent(self, parent):
        self.parent = parent

    @property
    def name(self):
        return self.element.name

    @property
    def label(self):
        return self.element.label

    @property
    def reference(self):
        return self.element.reference

    @property
    def depth(self):
        return len(self.get_parents())

    @property
    def path(self):
        parents = list(reversed(self.get_parents()))
        if len(parents) == 0:
            return self.name
        else:
            path = str(self.order) + " " + self.name
            for p in parents:
                path = p.name + "/" + path
            return path

    def get_parents(self):
        parents = []
        if self.parent is None:
            return parents
        else:
            p = self.parent
            while p is not None:
                parents.insert(0, p)
                p = p.parent
            return parents



import os
import shutil
import unittest
from xbrr.edinet.client.document_client import DocumentClient
from xbrr.edinet.reader.reader import Reader
from xbrr.edinet.reader.aspects.company import Company


class TestCompany(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        _dir = os.path.join(os.path.dirname(__file__), "../../data")
        client = DocumentClient()
        file_path = client.get_xbrl("S100G70J", save_dir=_dir,
                                    expand_level="dir")
        cls.reader = Reader(file_path)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.reader.xbrl_dir.root)
        if os.path.exists(cls.reader.taxonomy.root):
            shutil.rmtree(cls.reader.taxonomy.root)

    def test_history(self):
        feature = self.reader.extract(Company).history
        self.assertTrue(feature.normalized_text.startswith("2【沿革】"))

    def test_business_description(self):
        feature = self.reader.extract(Company).business_description
        self.assertTrue(feature.normalized_text.startswith("3【事業の内容】"))

    def test_affiliated_entities(self):
        feature = self.reader.extract(Company).affiliated_entities
        self.assertTrue(feature.normalized_text.startswith("4【関係会社の状況】"))

    def test_employees(self):
        feature = self.reader.extract(Company).employees
        self.assertTrue(feature.normalized_text.startswith("5【従業員の状況】"))



from pathlib import Path
from zipfile import ZipFile
import requests


class Taxonomy():
    TAXONOMIES = {
        "2013": "https://www.fsa.go.jp/search/20130821/editaxonomy2013New.zip",
        "2014": "https://www.fsa.go.jp/search/20140310/1c.zip",
        "2015": "https://www.fsa.go.jp/search/20150310/1c.zip",
        "2016": "https://www.fsa.go.jp/search/20160314/1c.zip",
        "2017": "https://www.fsa.go.jp/search/20170228/1c.zip",
        "2018": "https://www.fsa.go.jp/search/20180228/1c_Taxonomy.zip",
        "2019": "https://www.fsa.go.jp/search/20190228/1c_Taxonomy.zip",
        "2019_cg_ifrs": "https://www.fsa.go.jp/search/20180316/1c_Taxonomy.zip"
    }

    def __init__(self, taxonomy_root):
        self.root = taxonomy_root
        self.prefix = "http://disclosure.edinet-fsa.go.jp/taxonomy/"

    def __reduce_ex__(self, proto):
        return type(self), (self.root,)

    def download(self, year):
        year = str(year)
        expand_dir = self.root.joinpath("taxonomy").joinpath(year)
        taxonomy_file = self.root.joinpath(f"{year}_taxonomy.zip")

        download = False

        if not self.root.exists():
            self.root.mkdir(parents=True, exist_ok=True)
            download = True

        if not expand_dir.exists():
            expand_dir.mkdir(parents=True, exist_ok=True)
            download = True

        if download:
            # Download
            r = requests.get(self.TAXONOMIES[year], stream=True)
            with taxonomy_file.open(mode="wb") as f:
                for chunk in r.iter_content(1024):
                    f.write(chunk)

            # Extract
            with ZipFile(taxonomy_file, "r") as zip:
                for f in zip.namelist():
                    dirs = Path(f).parts
                    # Avoid Japanese path
                    taxonomy_at = dirs.index("taxonomy") if "taxonomy" in dirs else -1
                    if taxonomy_at > 0 and len(dirs) > (taxonomy_at + 1):
                        dirs = dirs[(dirs.index("taxonomy") + 1):]
                        _to = expand_dir.joinpath("/".join(dirs))
                        info = zip.getinfo(f)
                        if info.is_dir() and not _to.exists():
                            _to.mkdir(parents=True, exist_ok=True)
                        else:
                            _to.parent.mkdir(parents=True, exist_ok=True)
                            with _to.open("wb") as _to_f:
                                _to_f.write(zip.read(f))

            taxonomy_file.unlink()

        return expand_dir



import os
from xbrr.base.reader.base_element import BaseElement
from xbrr.edinet.reader.element_value import ElementValue


class Element(BaseElement):

    def __init__(self, name, element, reference, reader):
        super().__init__(name, element, reference, reader)
        self.name = name
        self.element = element
        self.reference = reference
        self.reader = reader

    @property
    def xsd(self):
        name = self.reference_name
        path = self.reference_path
        xml = None

        if path.endswith(".xsd"):
            xml = self.reader._read_from_cache(path)
        else:
            path = self._find_file(os.path.dirname(path), ".xsd")

        if os.path.dirname(path).endswith("PublicDoc"):
            element = xml.find("element", {"id": name})
        else:
            element = xml.find("xsd:element", {"id": name})

        return element

    def label(self, kind="lab", verbose=False):
        label = ""
        if kind == "en":
            label = self._get_label("_lab_en.xml", verbose)
        elif kind == "gla":
            label = self._get_label("_lab_gla.xml", verbose)
        else:
            label = self._get_label("_lab.xml", verbose)

        if label is None:
            return ""
        elif isinstance(label, str):
            return label
        else:
            return label.text

    def value(self, label_kind="", label_verbose=False):
        return ElementValue.create_from_element(
                reader=self.reader, element=self,
                label_kind=label_kind, label_verbose=label_verbose)

    def _get_label(self, extention, verbose):
        name = self.reference_name
        path = self.reference_path
        xml = None

        if os.path.dirname(path).endswith("PublicDoc"):
            label_path = self._find_file(os.path.dirname(path), extention)
            href = self.reference
        else:
            _dir = os.path.join(os.path.dirname(path), "label")
            label_path = self._find_file(_dir, extention)
            href = f"../{os.path.basename(path)}#{name}"

        xml = self.reader._read_from_cache(label_path)

        targets = self._read_link(
            xml=xml, arc_name="link:labelArc", reference=href,
            target_name="link:label", target_attribute="id")

        if len(targets) > 1:
            for lb in targets:
                if lb["xlink:role"].endswith("verboseLabel") and verbose:
                    label = lb
                    break
                else:
                    label = lb

        elif len(targets) > 0:
            label = targets[0]
        else:
            label = None

        return label

    def _read_link(self, xml, arc_name, reference="",
                   target_name="", target_attribute=""):

        # link: href: absolute path to element definition by url format.
        # name: underscore separated name. when used by tag, it is splited by ":"
        # name is solved by namespace so
        # name => link is good approach.

        reference = reference if reference else self.reference
        label = xml.find("link:loc", {"xlink:href": reference})
        arc = None

        if label is not None:
            arc = xml.find(arc_name, {"xlink:from": label["xlink:label"]})
        else:
            arc = xml.find(arc_name, {"xlink:label": self.name})

        if arc is None:
            return []

        target_name = target_name if target_name else "link:loc"
        target_attribute = target_attribute if target_attribute else "xlink:label"
        targets = []
        if arc is not None:
            targets = xml.find_all(target_name, {target_attribute: arc["xlink:to"]})
            if len(targets) == 0 and target_attribute != "xlink:label":
                targets = xml.find_all(target_name, {"xlink:label": arc["xlink:to"]})

        return targets



import os
import shutil
from bs4 import BeautifulSoup


class Directory():

    def __init__(self, root):
        self.root = root
        self._document_folder = os.path.join(root, "XBRL/PublicDoc")
        self._cache = {}

    def delete(self):
        shutil.rmtree(self.root)

    def _find_file(self, kind, as_xml=True):
        path = ""
        hit = False
        for f in os.listdir(self._document_folder):
            path = os.path.join(self._document_folder, f)
            if not os.path.isfile(path):
                path = ""
                continue

            if kind == "xbrl" and path.endswith(".xbrl"):
                hit = True
            elif kind == "xsd" and path.endswith(".xsd"):
                hit = True
            elif kind == "cal" and path.endswith("_cal.xml"):
                hit = True
            elif kind == "def" and path.endswith("_def.xml"):
                hit = True
            elif kind == "lab" and path.endswith("_lab.xml"):
                hit = True
            elif kind == "lab-en" and path.endswith("_lab-en.xml"):
                hit = True
            elif kind == "pre" and path.endswith("_pre.xml"):
                hit = True
            elif kind == "man" and f.startswith("manifest_") and f.endswith(".xml"):
                hit = True

            if hit:
                break

        if hit:
            if as_xml:
                xml = None
                if path in self._cache:
                    xml = self._cache[path]
                else:
                    with open(path, encoding="utf-8-sig") as f:
                        xml = BeautifulSoup(f, "lxml-xml")
                return xml
            else:
                return path
        else:
            return None

    @property
    def xbrl(self):
        return self._find_file("xbrl")

    @property
    def xsd(self):
        return self._find_file("xsd")

    @property
    def cal(self):
        return self._find_file("cal")

    @property
    def def_(self):
        return self._find_file("def")

    @property
    def lab(self):
        return self._find_file("lab")

    @property
    def lab_en(self):
        return self._find_file("lab-en")

    @property
    def pre(self):
        return self._find_file("pre")

    @property
    def man(self):
        return self._find_file("man")



class BaseClient():
    """"Base API Client.

    Manage the URL and version for the all client.
    """
    BASE_URL = "https://disclosure.edinet-fsa.go.jp/api/{}/{}"

    def __init__(self, target: str, version: str = "v1"):
        """
        Arguments:
            target -- API destination (set by subclass).

        Keyword Arguments:
            version {str} -- API version. (default: {"v1"}).
        """
        self.version = version
        self.target = target

    @property
    def endpoint(self):
        return self.BASE_URL.format(self.version, self.target)



import os
import shutil
import unittest
from xbrr.edinet.client.document_client import DocumentClient
from tests.utils import delay


class TestDocumentClient(unittest.TestCase):

    @delay
    def test_get(self):
        _dir = os.path.dirname(__file__)
        client = DocumentClient()
        file_path = client.get("S100FGR9", response_type=1, save_dir=_dir)
        self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)

    @delay
    def test_get_pdf(self):
        _dir = os.path.dirname(__file__)
        client = DocumentClient()
        file_path = client.get_pdf("S100FGR9", save_dir=_dir)
        self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)

    @delay
    def test_get_xbrl(self):
        _dir = os.path.dirname(__file__)
        client = DocumentClient()
        # lang="en" does not exist?
        file_path = client.get_xbrl("S100FGR9", save_dir=_dir)
        self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)

    @delay
    def test_get_xbrl_zip(self):
        _dir = os.path.dirname(__file__)
        client = DocumentClient()
        # lang="en" does not exist?
        file_path = client.get_xbrl("S100FGR9", save_dir=_dir,
                                    expand_level="dir")
        self.assertTrue(os.path.exists(file_path))
        self.assertTrue(os.path.isdir(file_path))
        shutil.rmtree(file_path)

    @delay
    def test_get_xbrl_zip_dir(self):
        _dir = os.path.dirname(__file__)
        client = DocumentClient()
        # lang="en" does not exist?
        file_path = client.get_xbrl("S100FGSC", save_dir=_dir,
                                    expand_level=None)
        self.assertTrue(os.path.exists(file_path))
        self.assertTrue(str(file_path).endswith(".zip"))
        os.remove(file_path)

    @delay
    def test_get_pdf_without_dir(self):
        client = DocumentClient()
        file_path = client.get_pdf("S100FGR9")
        self.assertTrue(os.path.exists(file_path))
        name = os.path.basename(file_path)
        self.assertTrue(str(name).startswith("S100FGR9_2__"))
        self.assertTrue(str(name).endswith(".pdf"))
        os.remove(file_path)



import importlib


class BaseReader():
    """
    Document to Element
    """

    def __init__(self, package):
        self.package = package

    def link_to_path(self, link):
        raise NotImplementedError("You have to implement link_to_path method.")

    def extract(self, aspect_cls_or_str, property=""):
        if not isinstance(aspect_cls_or_str, str):
            aspect_cls = aspect_cls_or_str
            return aspect_cls(self)

        aspect_str = aspect_cls_or_str
        imports = (
            "xbrr",
            self.package,
            "reader",
            "aspects",
            aspect_str
        )

        _class = None
        try:
            module = importlib.import_module(".".join(imports))

            def to_camel(snake_str):
                components = snake_str.split("_")
                return "".join(x.title() for x in components)

            _class_name = to_camel(aspect_str)
            _class = getattr(module, _class_name)

        except Exception as ex:
            raise Exception(f"Can't load class that matches {aspect_str} \n {ex}.")

        aspect = _class(self)
        feature = getattr(aspect, property)

        return feature



from xbrr.base.reader.base_parser import BaseParser
from xbrr.edinet.reader.element_value import ElementValue


class Business(BaseParser):

    def __init__(self, reader):
        tags = {
            "policy_environment_issue_etc": "jpcrp_cor:BusinessPolicyBusinessEnvironmentIssuesToAddressEtcTextBlock",
            "risks": "jpcrp_cor:BusinessRisksTextBlock",
            "research_and_development": "jpcrp_cor:ResearchAndDevelopmentActivitiesTextBlock",
            "management_analysis": "jpcrp_cor:ManagementAnalysisOfFinancialPositionOperatingResultsAndCashFlowsTextBlock",
            "overview_of_result": "jpcrp_cor:OverviewOfBusinessResultsTextBlock",
            "overview_of_value_chain": "jpcrp_cor:OverviewOfProductionOrdersReceivedAndSalesTextBlock",
            "analysis_of_finance": "jpcrp_cor:AnalysisOfFinancialPositionOperatingResultsAndCashFlowsTextBlock"
        }
        super().__init__(reader, ElementValue, tags)

    @property
    def policy_environment_issue_etc(self):
        return self.get_text_value("policy_environment_issue_etc")

    @property
    def risks(self):
        return self.get_text_value("risks")

    @property
    def management_analysis(self):
        return self.get_text_value("management_analysis")

    @property
    def research_and_development(self):
        return self.get_text_value("research_and_development")

    @property
    def overview_of_result(self):
        return self.get_text_value("overview_of_result")

    @property
    def overview_of_value_chain(self):
        return self.get_text_value("overview_of_value_chain")

    @property
    def analysis_of_finance(self):
        return self.get_text_value("analysis_of_finance")



import os
import unittest
from xbrr.edinet.reader.reader import Reader
from xbrr.edinet.reader.aspects.business import Business


class TestBusiness(unittest.TestCase):

    def get_xbrl(self, year=2019):
        path = os.path.join(os.path.dirname(__file__),
                            f"../../data/xbrl{year}.xbrl")
        xbrl = Reader(path)
        return xbrl

    def test_policy_environment_issue_etc(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Business).policy_environment_issue_etc
        self.assertTrue(feature.normalized_text.startswith("1【経営方針、経営環境及び対処すべき課題等】"))

    def test_risks(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Business).risks
        self.assertTrue(feature.normalized_text.startswith("2【事業等のリスク】"))

    def test_management_analysis(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Business).management_analysis
        self.assertTrue(feature.normalized_text.startswith("3【経営者による財政状態、経営成績及びキャッシュ・フローの状況の分析】"))

    def test_research_and_development(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Business).research_and_development
        self.assertTrue(feature.normalized_text.startswith("5【研究開発活動】"))

    def test_overview_of_result(self):
        xbrl = self.get_xbrl(2018)
        feature = xbrl.extract(Business).overview_of_result
        self.assertTrue(feature.normalized_text.startswith("1【業績等の概要】"))

    def test_overview_of_value_chain(self):
        xbrl = self.get_xbrl(2018)
        feature = xbrl.extract(Business).overview_of_value_chain
        self.assertTrue(feature.normalized_text.startswith("2【生産、受注及び販売の状況】"))

    def test_analysis_of_finance(self):
        xbrl = self.get_xbrl(2018)
        feature = xbrl.extract(Business).analysis_of_finance
        self.assertTrue(feature.normalized_text.startswith("7【財政状態、経営成績及びキャッシュ・フローの状況の分析】"))



import re
import unicodedata


class BaseParser():
    """
    Element to Value
    """

    def __init__(self, reader, value_class, tags=()):
        self.reader = reader
        self.value_class = value_class
        self.tags = {}
        if len(tags) > 0:
            self.tags = tags
        self._cache = {}

    def normalize(self, text):
        if text is None:
            return ""
        _text = text.strip()
        _text = unicodedata.normalize("NFKC", _text)
        return _text

    def read(self, name):
        if name not in self._cache:
            tag = self.tags[name]
            element = self.reader.find(tag)
            if element:
                self._cache[name] = element
            else:
                without_ns = tag.split(":")[1]  # without ns
                element = self.reader.find(without_ns)
                self._cache[name] = element

        return self._cache[name]

    def get_text_value(self, name):
        element = self.read(name)
        if element:
            value = self.value_class.create_from_element(
                        self.reader, element)

            html = element.html
            if html:
                value.value = html.text
            else:
                value.value = ""

            return value
        else:
            return self.value_class(self.tags[name])

    def search(self, name, pattern):
        element = self.read(name)
        ptn = re.compile(pattern)
        tags = element.html.find_all(["p", "span"])
        text = ""
        if tags and len(tags) > 0:
            for e in tags:
                _text = self.normalize(e.text)
                match = re.search(ptn, _text)
                if match:
                    text = _text
                    break

        return text

    def extract_value(self, name, prefix="", suffix="",
                      filter_pattern=None):
        element = self.read(name)
        text = element.html.text
        if filter_pattern is not None:
            text = self.search(name, filter_pattern)

        pattern = re.compile(f"({prefix}).+?({suffix})")
        match = re.search(pattern, text)
        value = ""

        if match:
            matched = match[0]
            value = matched.replace(prefix, "").replace(suffix, "")
            value = value.strip()
            if value.isdigit():
                value = int(value)
            elif value.replace(".", "").replace("．", "").isdigit():
                value = float(value)

        return value



import unittest
from datetime import datetime
from xbrr.edinet.client.document_list_client import MetaDataClient
from xbrr.edinet.client.document_list_client import DocumentListClient
from tests.utils import delay


class TestDocumentListClient(unittest.TestCase):

    @delay
    def test_metadata(self):
        client = MetaDataClient()
        metadata = client.get("2019-01-31")
        self.assertGreater(metadata.count, 0)

    @delay
    def test_metadata_by_datetime(self):
        client = MetaDataClient()
        date = datetime(2019, 3, 1)
        metadata = client.get(date)
        self.assertGreater(metadata.count, 0)

    @delay
    def test_document_list(self):
        client = DocumentListClient()
        documents = client.get("2019-01-31")
        self.assertEqual(documents.metadata.count, len(documents.list))

    @delay
    def test_document_list_by_datetime(self):
        client = DocumentListClient()
        date = datetime(2019, 3, 1)
        documents = client.get(date)
        self.assertEqual(documents.metadata.count, len(documents.list))



import unicodedata


class BaseElementValue():

    def __init__(self):
        pass

    def normalize(self, text):
        if text is None:
            return ""
        _text = text.strip()
        _text = unicodedata.normalize("NFKC", _text)
        return _text

    def to_dict(self):
        raise NotImplementedError(
            "You have to implement to_dict in subclass.")



import os
import unittest
from xbrr.edinet.reader.reader import Reader
from xbrr.edinet.reader.aspects.information import Information


class TestInformation(unittest.TestCase):

    def get_xbrl(self):
        path = os.path.join(os.path.dirname(__file__),
                            "../../data/xbrl2019.xbrl")
        xbrl = Reader(path)
        return xbrl

    def test_shareholders(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).shareholders
        self.assertTrue(feature.normalized_text.startswith("(5)【所有者別状況】"))

    def test_dividend_policy(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).dividend_policy
        self.assertTrue(feature.normalized_text.startswith("3【配当政策】"))

    def test_directors(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).directors
        self.assertTrue(feature.normalized_text.startswith("5【役員の状況】"))

    def test_corporate_governance(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).corporate_governance
        self.assertTrue(feature.normalized_text.startswith("(1)【コーポレート・ガバナンスの状況】"))

    def test_number_of_executives(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).number_of_directors
        self.assertEqual(feature.value, 14)

    def test_number_of_female_executives(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Information).number_of_female_executives
        self.assertEqual(feature.value, 1)



from xbrr.base.reader.base_parser import BaseParser
from xbrr.edinet.reader.element_value import ElementValue


class Information(BaseParser):

    def __init__(self, reader):
        tags = {
            "shareholders": "jpcrp_cor:ShareholdingByShareholderCategoryTextBlock",
            "dividend_policy": "jpcrp_cor:DividendPolicyTextBlock",
            "directors": "jpcrp_cor:InformationAboutOfficersTextBlock",
            "corporate_governance": "jpcrp_cor:ExplanationAboutCorporateGovernanceTextBlock"
        }
        super().__init__(reader, ElementValue, tags)

    @property
    def shareholders(self):
        return self.get_text_value("shareholders")

    @property
    def dividend_policy(self):
        return self.get_text_value("dividend_policy")

    @property
    def directors(self):
        return self.get_text_value("directors")

    @property
    def corporate_governance(self):
        return self.get_text_value("corporate_governance")

    @property
    def number_of_directors(self):
        value = self.get_text_value("directors")
        numbers, ground = self._extract_number_of_directors()
        value.value = numbers["total"]
        value.ground = ground
        return value

    @property
    def number_of_female_executives(self):
        value = self.get_text_value("directors")
        numbers, ground = self._extract_number_of_directors()
        value.value = numbers["female"]
        value.ground = ground
        return value

    def _extract_number_of_directors(self):
        text = self.search("directors", "^(男性).+(名).+(女性).+(名)")
        numbers = {
            "male": 0,
            "female": 0,
            "total": 0,
        }

        total = 0
        for p, s in [("男性", "名"), ("女性", "名")]:
            value = self.extract_value("directors", p, s)
            if isinstance(value, int):
                if p == "男性":
                    numbers["male"] = value
                elif p == "女性":
                    numbers["female"] = value
                total += value

        numbers["total"] = total
        return numbers, text



import os
import shutil
import unittest
from xbrr.edinet.client.document_client import DocumentClient
from xbrr.edinet.reader.reader import Reader
from xbrr.edinet.reader.aspects.finance import Finance


class TestFinance(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        _dir = os.path.join(os.path.dirname(__file__), "../../data")
        client = DocumentClient()
        file_path = client.get_xbrl("S100G6IU", save_dir=_dir,
                                    expand_level="dir")
        cls.reader = Reader(file_path)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.reader.xbrl_dir.root)
        if os.path.exists(cls.reader.taxonomy.root):
            shutil.rmtree(cls.reader.taxonomy.root)

    def get_xbrl(self):
        path = os.path.join(os.path.dirname(__file__),
                            "../../data/xbrl2019.xbrl")
        xbrl = Reader(path)
        return xbrl

    def test_voluntary_accounting_policy_change(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Finance).voluntary_accounting_policy_change

    def test_segment_information(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Finance).segment_information
        self.assertTrue(feature.normalized_text.startswith("(セグメント情報等)"))

    def test_real_estate_for_lease(self):
        xbrl = self.get_xbrl()
        feature = xbrl.extract(Finance).real_estate_for_lease

    def test_bs(self):
        bs = self.reader.extract(Finance).bs()
        # bs.to_csv("bs.csv", index=False, encoding="shift_jis")
        self.assertTrue(bs is not None)
        self.assertGreater(len(bs), 0)

    def test_pl(self):
        pl = self.reader.extract(Finance).pl()
        # pl.to_csv("pl.csv", index=False, encoding="shift_jis")
        self.assertTrue(pl is not None)
        self.assertGreater(len(pl), 0)

    def test_bs_ifrs(self):
        bs = self.reader.extract(Finance).bs(ifrs=True, link_type="presentation")
        # bs.to_csv("bs_ifrs.csv", index=False, encoding="shift_jis")
        self.assertTrue(bs is not None)
        self.assertGreater(len(bs), 0)

    def test_pl_ifrs(self):
        pl = self.reader.extract(Finance).pl(ifrs=True)
        # pl.to_csv("pl_ifrs.csv", index=False, encoding="shift_jis")
        self.assertTrue(pl is not None)
        self.assertGreater(len(pl), 0)



from xbrr.base.reader.base_parser import BaseParser
from xbrr.edinet.reader.element_value import ElementValue


class Company(BaseParser):

    def __init__(self, reader):
        tags = {
            "history": "jpcrp_cor:CompanyHistoryTextBlock",
            "business_description": "jpcrp_cor:DescriptionOfBusinessTextBlock",
            "affiliated_entities": "jpcrp_cor:OverviewOfAffiliatedEntitiesTextBlock",
            "employees": "jpcrp_cor:InformationAboutEmployeesTextBlock"
        }
        super().__init__(reader, ElementValue, tags)

    @property
    def history(self):
        return self.get_text_value("history")

    @property
    def business_description(self):
        return self.get_text_value("business_description")

    @property
    def affiliated_entities(self):
        return self.get_text_value("affiliated_entities")

    @property
    def employees(self):
        return self.get_text_value("employees")
